/***********************************************************************
 * Time-stamp: <Wednesday, 18.07.2001 at 19:16:43; edited by bauke>
 * Testing random rumber generators
 ***********************************************************************/

#include <iostream>
#include <cstdlib>
#include <new>
#include <cmath>
#include <vector>
#include "trng.h"

using namespace TRNG;
using namespace std;


// for special functions calculation see
// Samuel S. M. Wong Computational Methods in Physics and Engineering


double TRNG::Gamma(double x) {
  // Gamma function for positive argument
  //
  //               infinity
  //              /
  //             |           (x - 1)
  //             |          t        exp(-t) dt
  //             |
  //            /
  //              0
  double f, z, t;
  if (x<0.0)
    throw error("invalid (negative) argument for TRNG::Gamma");
  // use a interpolating polynom for 1<=x<2
  if (x<15.0) {
    // Gamma(x)=Gamma(x+1)/x
    if (x<1.0) {
      f=1.0/x;
      z=x;
    }
    if (1.0<=x && x<2.0) {
      z=x-1.0;
      f=1.0;
    }
    // Gamma(x+1)=x*Gamma(x)
    if (x>=2.0) {
      f=1.0;
      z=x-1.0;
      do {
	f=f*z;
	z--;
      } while (z>=1.0);
    }
    return(f*(1.0+
	      (-0.5772152952998961+
	       (0.9890418621994901+
		(-0.9072503205615771+
		 (0.9796139097861202+
		  (-0.9693649942410396+
		   (0.9409502806615237+
		    (-0.8401327394958836+
		     (0.6506941079199484+
		      (-0.4028899587819087+
		       (0.1812400864733248+
			(-0.5156307405048575E-1
			 +0.6876141229E-2
			 *z)*z)*z)*z)*z)*z)*z)*z)*z)*z)*z)*z));
  }
  // Stirling's approximation
  t=x*x;
  return(exp(-x)*pow(x, x-0.5)*2.5066282746310*
	 (1.0
	  +1.0/12.0/x
	  +1/288.0/t
	  -139.0/51840.0/(t*x)
	  -571.0/2488320.0/(t*t)));
}


double TRNG::GammaP(double a, double x) {
  // incomplete Gamma function
  //
  //                  x
  //                 /
  //                |            (a - 1)
  //                |   exp(-t) t        dt
  //                |
  //               /
  //                 0
  //               ------------------------
  //                       Gamma(a)
  double t;
  t=Gamma(a);
  if (x<a+1.0)
    return(Gamma_ser(a, x)/t);
  else
    return(1.0-Gamma_cf(a, x)/t);
}


double TRNG::GammaQ(double a, double x) {
  // incomplete Gamma function
  //
  //               infinity
  //              /
  //             |                   (a - 1)
  //             |          exp(-t) t        dt
  //             |
  //            /
  //              x
  //            -------------------------------
  //                       Gamma(a)
  double t;
  t=Gamma(a);
  if (x<a+1.0)
    return(1.0-Gamma_ser(a, x)/t);
  else
    return(Gamma_cf(a, x)/t);
}


double TRNG::incomp_Gamma(double a, double x) {
  if (x<a+1.0)
    return(Gamma_ser(a, x));
  else
    return(Gamma(a)-Gamma_cf(a, x));
}


double TRNG::comp_incomp_Gamma(double a, double x) {
  if (x<a+1.0)
    return(Gamma(a)-Gamma_ser(a, x));
  else
    return(Gamma_cf(a, x));
}


double TRNG::Gamma_ser(double a, double x) {
  // incomplete Gamma function's power series representation
  // for x<a+1
  const int itmax=150;
  const double eps=1e-12;
  int i;
  double sum, xx, n; 
  if (x<0.0)
    throw error("x less than 0 in routine TRNG::Gamma_ser");
  if (a<=0.0)
    throw error("a less or equal than 0 in routine TRNG::Gamma_ser");
  if (x==0.0)
    return(0.0);
  xx=1.0/a;
  n=a;
  sum=xx;
  i=0;
  do {
    ++n;
    ++i;
    xx*=x/n;
    sum+=xx;
  } while (fabs(xx)>eps*fabs(sum) && i<itmax);
  if (i==itmax)
    throw error("convergence problem in TRNG::Gamma_ser");
  return(exp(-x+a*log(x))*sum);
}


double TRNG::Gamma_cf(double a, double x) {
  // complementary incomplete Gamma function's continued fraction  
  // representation for x>a+1
  const double itmax=75.0;
  const double eps=1.0e-12;
  const double min=1.0e-45;
  double ai, bi, ci, di, del, h, i; 
  // Set up for evaluating continued fraction by modied Lentz's method
  bi=x+1.0-a; 
  ci=1.0/min; 
  di=1.0/bi; 
  h=di;
  i=0.0;
  do { 
    // Iterate to convergence. 
    ++i;
    ai=-i*(i-a); 
    bi+=2.0;
    di=ai*di+bi;
    if (fabs(di)<min) 
      di=min; 
    ci=bi+ai/ci; 
    if (fabs(ci)<min)
      ci=min;
    di=1.0/di;
    del=di*ci;
    h*=del;
  } while ((fabs(del-1.0)>eps) && i<itmax);
  if (i==itmax)
    throw error("a too large or convergence problem in TRNG::Gamma_cf"); 
  return(exp(-x+a*log(x))*h);
}


double TRNG::errf(double x) {
  // error function
  //
  //                  x
  //               /
  //              |                     2
  //              |          exp(- 1/2 t ) dt
  //              |
  //             /
  //               -infinity
  //             ----------------------------
  //                      sqrt(2 Pi)
  return((x>0.0) ? 0.5*GammaP(0.5, 0.5*x*x) : -0.5*GammaP(0.5, 0.5*x*x));
}


double TRNG::Chisquare(vector<double> &prob, vector<double> &observ) {
  vector<double>::iterator i, j;
  double Chi2, n, t1, t2;
  if (prob.size()!=observ.size())
    throw error("different sized vectors in TRNG::Chisquare");
  n=0.0;
  Chi2=0.0;
  for (i=prob.begin(), j=observ.begin(); 
       i!=prob.end() && j!=observ.end(); ++i, ++j) {
    n+=*j;
    if ((*j)<5.0)
      throw error("TRNG::Chisquare: not enough observations");
  }
  for (i=prob.begin(), j=observ.begin(); 
       i!=prob.end() && j!=observ.end(); ++i, ++j) {
    t1=(*i)*n;
    t2=(*j)-t1;
    Chi2+=t2*t2/t1;
  }
  return(Chi2);
}


double TRNG::Chisquare_prob(double Chi2, long df) {
  double ddf;
  ddf=static_cast<double>(df);
  if (Chi2==0.0)
    return(1.0);
  if (df<256l)
    return(GammaQ(0.5*ddf, 0.5*Chi2));
  // see Knuth, The Art of Computer Programming II, page 41
  return(0.5-errf(-3.0/4.0*sqrt(2.0)*sqrt(ddf)+
		  1.0/4.0*sqrt(-6*ddf+24*Chi2+16)));
}


double TRNG::Stirling_num2(long n, long m) {
  // see Knuth, The Art of Computer Programming I
  if (n<0l || m<0l)
    throw error("negative argument for TRNG::Stirling_num2");
  if (m>n)
    return(0.0);
  if (m==n)
    return(1.0);
  if (m==0l)
    return(0.0);
  if (m==1l)
    return(1.0);
  if (m==2l)
    return(static_cast<double>((1ll<<(n-1l))-1ll));
  return(m*Stirling_num2(n-1l, m)+Stirling_num2(n-1l, m-1l));
}


double TRNG::Student_t(double p, long ml, bool symmetric) {
  const double pi=3.14159265358979324;
  const double sqrt2=1.41421356237309505;
  const double two_div_sqrt_Pi=1.12837916709551257;
  double t, dt, t2, g1, g2, g3, g4, m;
  int i;
  bool less_0_5;
  if (ml<1l)
    throw error("less than one degree of freedom in TRNG::Student_t");
  if (p<=0.0 || p>=1.0)
    throw error("probability out of range in TRNG::Student_t");
  m=static_cast<double>(ml);
  if (symmetric)
    p+=0.5*(1.0-p);
  // exact formulas from Statistik. Lehr- und Handbuch der angewandten 
  // Statistik., Joachim Hartung, Brbel Elpelt, Karl-Heinz Klsener 
  // R. Oldenbourg Verlag, 1998, page 892
  if (ml==1l) {
    t=tan(pi*(p-0.5));
  }
  if (ml==2l) {
    t=(2.0*p-1.0);
    t*=sqrt2/sqrt(1-t*t); 
  }
  if (ml>2l) {
    // use symmetry t(1-p, m)=-t(p, m)
    less_0_5=(p<0.5);
    if (less_0_5)
      p=1.0-p;
    // claculate t(p, m) for m=oo with Newton's method 
    t=0.0;
    i=0;
    do {
      dt=-(2.0*errf(sqrt2*t)+1.0-2.0*p)/(two_div_sqrt_Pi*exp(-t*t));
      t+=dt;
    } while (abs(dt/t)>1e-14 && ++i<12);
    // use approximation form Handbook of Mathematical Functions, 
    // With Formulas, Graphs, and Mathematical Tables, Milton Abramowitz 
    // Dover Publications, Inc., page 949
    t*=sqrt2;
    t2=t*t;
    g1=1.0/4.0*(1.0+t2)*t;
    g2=1.0/96.0*(3.0+(16.0+5.0*t2)*t2)*t;
    g3=1.0/384.0*(-15.0+(17.0+.0*(19.0+3.0*t2)*t2)*t2)*t;
    g4=1.0/92160.0*(-945.0+(-1920.0+(1482.0+(776.0+79*t2)*t2)*t2)*t2)*t;
    t+=g1/m+g2/m/m+g3/m/m/m+g4/m/m/m/m;
    if (less_0_5)
      t*=-1.0;
  }
  return(t);
}
